package fi.helsinki.w3m.metadata;

import java.net.URL;
import java.util.*;
import java.io.*;

import com.hp.hpl.jena.rdf.query.util.ModelLoader;
import com.hp.hpl.mesa.rdf.jena.model.* ;
import com.hp.hpl.mesa.rdf.jena.mem.*;
import com.hp.hpl.mesa.rdf.jena.common.* ;
import com.hp.hpl.jena.util.* ;

import fi.helsinki.w3m.metadata.ontology.*;
import fi.helsinki.w3m.metadata.item.*;

/*
 * $Revision: 1.86 $
 * $Date: 2002/05/17 12:31:19 $
 */

/**
 *An object model wrapping Jena.
 */
public class MuseumModel {

	/* Magic strings, TODO: read from config file! */
	public static final String TYPE_PROPERTY_URI = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type";
	public static final String LABEL_PROPERTY_URI = "http://www.w3.org/TR/1999/PR-rdf-schema-19990303#label";
	public static final String SUBCLASSOF_PROPERTY_URI = "http://www.w3.org/TR/1999/PR-rdf-schema-19990303#subClassOf";
	public static final String WEIGHT_ID_PROPERTY_URI = "http://protege.stanford.edu/klik#Id";
	public static final String WEIGHT_COUNT_PROPERTY_URI = "http://protege.stanford.edu/klik#count";
	public static final String RESOURCE_CLASS_URI = "http://www.w3.org/TR/1999/PR-rdf-schema-19990303#Resource";
	public static final String RESOURCE_WEIGHT_URI = "http://protege.stanford.edu/klik#resource_weight";

    public static final String BASE_NS = "http://protege.stanford.edu/Promootio#";	
    public static final String WEIGHT_NS = "http://protege.stanford.edu/klik#";
	public static final String IMFILE_NS="http://protege.stanford.edu/imagefile#";
	public static final String FACETNAME_PREDICATE = "http://protege.stanford.edu/fasetit#Nimi";
	public static final String INSTANCELABEL_PREDICATE = "http://protege.stanford.edu/Promootio#Nimi";
	
    public static final String NAME_PREFIX = "#";
    public static final String INSTANCE_PREFIX = "promootio_";	
    public static final String IMAGE_PREFIX = "img_";	
    public static final String WEIGHT_SUFFIX = "_weight";	
	public static final String IMAGE_PROTOCOL = "file://";
	public static final String DC_DESCRIPTION = "http://purl.org/metadata/dublin_core#Description";
	public static final String DC_SUBJECT = "http://purl.org/metadata/dublin_core#Subject";	

	public static final String IMAGE_PATH = "pictures/";
	public static final String RDF_PATH = "rdf/";
	public static final String ONTOLOGY_PATH = "rdf/ontology/";	
	
    /**
     *The base model this class wraps.
     */
    private Model baseModel;

    /**
     *A model holding the factes of the MuseumModel.
     */
    private Model facetModel;

    /**
     *A model holding the weights of the objects in the MuseumModel.
     */
    private Model weightModel;

    /**
     *A model holding the file info of the images in the MuseumModel.
     */
    private Model imagefileModel;

    /**
     *Path where to find the files related to the model
     */
    private String basePath;

    /**
     *Base URL for the application.
     */
    private java.net.URL baseURL;

	/**
	*Check for not loading the model twice.
	*/
	private boolean modelReadAlready;

    /**
     *An Item-object holding the access count of the most popular image.
     */
	private Item imageWeightMaximumItem;

    /**
     *An Item-object holding the access count of the most popular ontology class.
     */
	private Item classWeightMaximumItem;

    /**
     *An Item-object holding the access count of the most popular image property.
     */	
	private Item imagePropertyWeightMaximumItem;
	
	/**
	*A thread to save the model every hour.
	*/
	private class ModelSaver extends Thread {
		public void run() {
			while (!this.interrupted()) {
				try {				
					sleep(1000*60*60);
				} catch (InterruptedException ie) {
					System.out.println("Thread interrupted: "+ ie);	
					break;
				}
				writeModel();
			}
		/* If we were interrupted while writing, don't write again */	
		if (this.interrupted()) writeModel();		
		}
	}
	
	private ModelSaver saver = new ModelSaver();
		
    /**
     *Constructor for the object model
     *
     *@param baseURL Base URL for the application.
     *@param basePath Path where to find the files related to the model.
     */
    public MuseumModel(java.net.URL baseURL, String basePath) {
		this.baseURL = baseURL;
		if (basePath.endsWith("/")) {
			this.basePath = basePath;    
		} else {
			this.basePath = basePath + "/";    
		}
		modelReadAlready = false;
		imageWeightMaximumItem = new Item(this, new SimpleId("ImageWeightMaximumhack"));
	    classWeightMaximumItem = new Item(this, new SimpleId("ClassWeightMaximumhack"));
		imagePropertyWeightMaximumItem = new Item(this, new SimpleId("ImagePropertyWeightMaximumhack"));
    }

    /**
     *Creates the model by reading RDF documents from the disk.
     *
     *@return True if the reading succeeded.
     */
    public boolean readModel() {
		if (modelReadAlready) {
			return false;
		} else {
			modelReadAlready = true;
		}
		
		try {
		    /* Read the files containing the ontologies. */
		    baseModel = ModelLoader.loadModel(basePath + ONTOLOGY_PATH + "promootio.rdfs", ModelLoader.langXML);
	    	baseModel = baseModel.read("file:"+basePath + ONTOLOGY_PATH + "promootio.rdf");
			baseModel = baseModel.read("file:"+basePath + RDF_PATH + "img.rdf");
		
			imagefileModel = ModelLoader.loadModel(basePath + ONTOLOGY_PATH + "imagefile.rdfs", ModelLoader.langXML);
	    	imagefileModel = imagefileModel.read("file:"+basePath + ONTOLOGY_PATH + "imagefile.rdf");
		
		    facetModel = ModelLoader.loadModel(basePath + ONTOLOGY_PATH + "fasetit.rdfs", ModelLoader.langXML);
		    facetModel = facetModel.read("file:"+basePath + ONTOLOGY_PATH + "fasetit.rdf");
		
	    	weightModel = ModelLoader.loadModel(basePath + ONTOLOGY_PATH + "klik.rdfs", ModelLoader.langXML);
		    weightModel = weightModel.read("file:"+basePath + ONTOLOGY_PATH + "klik.rdf");
       
        } catch (Exception e) {
            return false;
        }
		saver.start();
		return true;
    }

    /**
     *Writes the weight- and imagefile-models to disk.
     *
     *@return True if the write succeeded.
     */
    public synchronized boolean writeModel() {
		if (modelReadAlready) { 
			try {
				System.out.print("Writing models.");					
		    	FileWriter weightWriter = new FileWriter(basePath + ONTOLOGY_PATH + "klik.rdf");	
 			    weightModel.write(weightWriter);

			    FileWriter imagefileWriter = new FileWriter(basePath + ONTOLOGY_PATH + "imagefile.rdf");	
			    imagefileModel.write(imagefileWriter);			
				System.out.println(".done.");
				return true;
			} catch (Exception E) {
				System.out.println(".failed!");
				System.out.println(E.toString());			
			}
		}
		return false;		
    }

    /**
     *Returns the facets of the main ontology according to the facet ontology.
     *
     *@return A ResourceCollection-object holding OntologyClass-objects of the facet roots.
     */
    public ResourceCollection getFacetRoots() throws RDFException {
		PropertyImpl facetNameProperty = new PropertyImpl(FACETNAME_PREDICATE);
		SelectorImpl facetNameSelector = new SelectorImpl(null, facetNameProperty, (String) null);        		
	
		final MuseumModel model = this;
		ResourceProvider rp = new ResourceProvider() {
			public OntologyResource getResource( Object o ) {
				/* We *know* this is an ontology class name so we create an id from it. */
				Statement stmt = (Statement) o;
				SimpleId oclassId = new SimpleId(encodeURI(BASE_NS+stmt.getObject().toString()));
			    return new OntologyClass(model, oclassId);
			}
		};
		StatementResourceCollection rv = new StatementResourceCollection(this.getFacetModel(), facetNameSelector, rp);
		return rv;
    }
	
    /**
     *Returns all images.
     *
     *@return A ResourceCollection-object holding Image-objects of all images.
     */
	public ResourceCollection getAllImages() {
		OntologyClass resourceClass = new OntologyClass(this, new SimpleId(RESOURCE_CLASS_URI));
		Iterator imageCiter = resourceClass.getImages(OntologyClass.ALL).iterator();
		return resourceClass.getImages(OntologyClass.ALL);
	}
	
    /**
     *Returns the (at maximum) n most popular images.  Specify a negative value to get all images. The images will be sorted in descending order according to their weight.
     *
     *@param numberOfImages The maximum number of images to return.
     *@return A ResourceCollection-object holding Image-objects of the images.
     */
    public ResourceCollection getPopularImages(int numberOfImages) {
		/*  This needs some optimisation, since we have *many* more images than facets. If we can decide that we alway want the
			same number of popular images, a hashtable would be the cleanest way to do this.. */
		try {
			Iterator imageIterator = this.getAllImages().iterator();
			LinkedList sortedImages = new LinkedList();
			while (imageIterator.hasNext()) {
				sortedImages.add(imageIterator.next());
			}
			/* Sort in descending order, return only the specified number of results or all of them if too many were specified */
			Collections.sort(sortedImages, new WeightComparator());

			if (numberOfImages < 0) {
				return new SimpleResourceCollection(sortedImages);
			} else {
				if (sortedImages.size() < numberOfImages) numberOfImages = sortedImages.size();				
				return new SimpleResourceCollection(sortedImages.subList(0, numberOfImages));
			}
		} catch (Exception E) {
			return null;
		}
    }

    /**
     *Returns the (at maximum) n most popular facets. Specify a negative value to get all facets. The facets will be sorted in descending order according to their weight.
     *
     *@param numberOfFacets The maximum number of facets to return.
     *@return A ResourceCollection-object holding OntologyClass-objects of the facets.
     */
    public ResourceCollection getPopularFacets(int numberOfFacets) {
		/* Probably ok to leave like this since we have very few facets */
		try {
			Iterator facetIterator = this.getFacetRoots().iterator();
			LinkedList sortedFacets = new LinkedList();
			while (facetIterator.hasNext()) {
				sortedFacets.add(facetIterator.next());
			}
			/* Sort in descending order, return only the specified number of results or all of them if too many were specified */
			Collections.sort(sortedFacets, new WeightComparator());

			if (numberOfFacets < 0) {
				return new SimpleResourceCollection(sortedFacets);
			} else {
				if (sortedFacets.size() < numberOfFacets) numberOfFacets = sortedFacets.size();				
				return new SimpleResourceCollection(sortedFacets.subList(0, numberOfFacets));
			}
		} catch (Exception E) {
			return null;
		}
    }

    /**
     *Returns an ontology class by its id.
     *
     *@param classId The id of the class to return.
     *@return An OntologyClass-object of the class.
     */
    public OntologyClass getOntologyClass(Id classId) {
		/* TODO: Check for existence */
		OntologyClass newClass = new OntologyClass(this, classId);

		return newClass;
    }

    /**
     *Returns an image class by its name.
     *
     *@param imageId The code of the image to return.
     *@return An Image-object of the image.
     */
    public Image getImage(Id imageId) {
		/* TODO: Check for existence */
		Image newImage = new Image(this, imageId);

		return newImage;
    }

    /**
     *Returns the weight of the most popular image property in the model.
     *
     *@return A long value holding the weight of the most popular image property.
     */
    public long getImagePropertyWeightMaximum() {
		return this.imagePropertyWeightMaximumItem.getWeight();
    }

    /**
     *Sets the weight of the most popular image property in the model.
     *
     *@param count A long value holding the new maximum weight value.
     *@return A long value holding the weight of the most popular image property.
     */
    public long setImagePropertyWeightMaximum(long count) {
	if (this.imagePropertyWeightMaximumItem.getWeight() < count) {
	    this.imagePropertyWeightMaximumItem.setWeight(count);
	}
	return this.imagePropertyWeightMaximumItem.getWeight();
    }

    /**
     *Returns the weight of the most popular image in the model.
     *
     *@return A long value holding the weight of the most popular image.
     */
    public long getImageWeightMaximum() {
		return this.imageWeightMaximumItem.getWeight();
    }

    /**
     *Sets the weight of the most popular image in the model.
     *
     *@param count A long value holding the new maximum weight value.
     *@return A long value holding the weight of the most popular image.
     */
    public long setImageWeightMaximum(long count) {
	if (this.imageWeightMaximumItem.getWeight() < count) {
	    this.imageWeightMaximumItem.setWeight(count);
	}
	return this.imageWeightMaximumItem.getWeight();
    }

    /**
     *Returns the weight of the most popular ontology class in the model.
     *
     *@return A long-value holding the weight of the most popular ontology class.
     */
    public long getClassWeightMaximum() {
		return this.classWeightMaximumItem.getWeight();
    }

    /**
     *Sets the weight of the most popular image in the model.
     *
     *@param count A long-value holding the new maximum weight value.
     *@return A long-value holding the weight of the most popular ontology class.
     */
    public long setClassWeightMaximum(long count) {
	if (this.classWeightMaximumItem.getWeight() < count) {
	    this.classWeightMaximumItem.setWeight(count);
	}
	return this.classWeightMaximumItem.getWeight();
    }

    /**
     *Returns the base path of the model.
     *
     *@return A String-object holding the base path of this model.
     */
    public String getBasePath() {
		return this.basePath;
    }

    /**
     *Returns the base URL of the model.
     *
     *@return An URL-object holding the base URL of this model.
     */
    public URL getBaseURL() {
		return this.baseURL;
    }

    /**
     *Returns the base model this class wraps.
     *
     *@return A Model-object holding the base model this class wraps.
     */
    public Model getBaseModel() {
		return this.baseModel;
    }

    /**
     *Returns the weight-model in this class.
     *
     *@return A Model-object holding the weight-model in this class.
     */
    public Model getWeightModel() {
		return this.weightModel;
    }

    /**
     *Returns the imagefile-model in this class.
     *
     *@return A Model-object holding the iamgefile-model in this class.
     */
    public Model getImagefileModel() {
		return this.imagefileModel;
    }

    /**
     *Returns the facet-model in this class.
     *
     *@return A Model-object holding the facet-model in this class.
     */
    public Model getFacetModel() {
		return this.facetModel;
    }

    /**
     *Wraps Jena's URIref.encode method. For details see http://www.w3.org/TR/charmod/#sec-URIs
     *
     *@param stringToEncode The string to encode.
     *@return The encoded string.
     */
    public static String encodeURI(String stringToEncode) {
		if (stringToEncode == null) {
			return "";
		} else {
			return URIref.encode(stringToEncode);
		}
    }

    /**
     *Wraps Jena's URIref.decode method. For details see http://www.w3.org/TR/charmod/#sec-URIs
     *
     *@param stringToEncode The string to decode.
     *@return The decoded string.
     */
    public static String decodeURI(String stringToDecode) {
		if (stringToDecode == null) {
			return "";	
		} else {
			return URIref.decode(stringToDecode);
		}
    }

	/** 
	*Test method for selectors. Prints all resulting statements to System.out
	*
	*@param	selector The selector to test.
	*@param model The jena model to test the selector in.
	*/
	public static void testSelector (SelectorImpl selector, Model model) {
		try {
	   	StmtIterator stmtIter = model.listStatements(selector);
        while (stmtIter.hasNext()) {
            Statement stmt = stmtIter.next();
			String statement = stmt.toString();
			System.out.println(statement);
		}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	*Write the model to disk when the model is destroyed.
	*/
	protected void finalize() throws Throwable {
	super.finalize();
	saver.interrupt();
	this.writeModel();
	}
}

/**
*A class that implemements the Comparator for comparing OntologyResources according to weight.
*/
class WeightComparator implements Comparator {
	public int compare(Object o1, Object o2) {	
		/* Compare the weight values and reverse the order for a decending order. */
		Long weight1 = new Long(((OntologyResource)o1).getWeight());
		Long weight2 = new Long(((OntologyResource)o2).getWeight());
		return weight2.compareTo(weight1);
	}
}
