package fi.helsinki.w3m.metadata;

import java.util.*;
import java.net.*;

import com.hp.hpl.mesa.rdf.jena.model.* ;
import com.hp.hpl.mesa.rdf.jena.common.* ;
import com.hp.hpl.jena.util.* ;

import fi.helsinki.w3m.metadata.ontology.*;
import fi.helsinki.w3m.metadata.item.*;

/*
 * $Revision: 1.27 $
 * $Date: 2002/05/17 12:31:19 $
 */

/**
 * This class provides an iterator for ResourceCollections.
 * Iterators of this type can be linked. 
 * 
 * @author Samppa Saarela
 */
public abstract class ResourceIterator implements Iterator{

    /**
     * Next element of this iterator. 
     */
    private OntologyResource next;

    /**
     * Iterator to be used after this iterator is empty. 
     */
    private ResourceIterator tail;

    /**
     * Used to construct OntologyResources. 
     */
    private ResourceProvider resourceProvider;

    /**
     * Is this iterator closed?
     */
    private boolean closed = false;


    /**
     * Constructs an empty iterator. 
     */
    protected ResourceIterator() {}

    /**
     * Constructs a new ResourceIterator that uses resourceProvider
     * to construct its elements.
     *
     * @param resourceProvider Used to constuct the elements of this iterator. 
     */
    protected ResourceIterator(ResourceProvider resourceProvider) {
	this.resourceProvider = resourceProvider;
    }

    /**
     * Adds a new iterator to the tail of this iterator. 
     *
     * @param resourceIterator Iterator to be added. 
     */
    public void addIterator(ResourceIterator resourceIterator) {
	if (!this.closed) {
	    if (this.tail == null) {
		this.tail = resourceIterator;
	    }
	    else {
		this.tail.addIterator(resourceIterator);
	    }
	}
    }

    /**
     * This method tries to figure out the next element, using <tt>hasNextHead()</tt>
     * and <tt>getNext()</tt>. If <tt>getNext()</tt> returns <tt>null</tt> it will be called again
     * until it returns not null object or <tt>hasNextHead()</tt> returns false.
     * After consuming this iterator method continues to consume tail - if there is one. 
     * This method set's the attribute <tt>next</tt>. If resourceProvider is null
     * the object is casted to OntologyResource.
     *
     * @returns true only if there actually exists a next element.
     */
    public boolean hasNext() {
	try {
	    if (!this.closed) {
		// While next is null, try to get a new (skip null's in case filtering is applied)
		while (this.next == null && hasNextHead()) {
		    if (this.resourceProvider != null) {
			this.next = this.resourceProvider.getResource(getNext());
		    }
		    // The user has not provided ResourceProvider so try casting...
		    else {
			try {
			    Object nextObject = getNext();
			    if (nextObject != null && nextObject instanceof OntologyResource) 
				this.next = (OntologyResource)nextObject;
			} catch (ClassCastException cce) {
			    cce.printStackTrace();
			}
		    }
		}
	    }
	    
	    // If this iterator is consumed, close it...
	    if (this.next == null) {
		close(true);
		// ...and try to get next from tail (if there is one)
		if (this.tail != null) {
		    // While next is null, try to get a new (skip null's in case filtering is applied)
		    while (this.next == null && this.tail.hasNext()) {
			this.next = (OntologyResource)this.tail.next();
		    }
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return this.next != null;
    }

    /**
     * This method returns <tt>next</tt>. If there isn't 
     * anymore elements left this method returns null and closes this
     * iterator and <tt>tail</tt>.
     * 
     * @returns next element or null if there isn't any more elements left.
     */
    public Object next() {
	OntologyResource rv = null;
	if (hasNext()) {
	    // Return next...
	    rv = this.next;
	    // ...so it's no longer next
	    this.next = null;
	}
	else {
	    close();
	}
	return rv;
    }


    /**
     * Returns the next element. Any filtering should be done here in the derived classes.
     */
    protected abstract Object getNext() throws Exception;

    /**
     * Checks if this iterator has more elements.
     */
    protected abstract boolean hasNextHead() throws Exception;


    /**
     * Closes this iterator by releasing any reserved resources.
     * If this iterator is actually a ResultSet 
     * from a database query it is very important to close it here.
     */
    protected abstract void closeHead() throws Exception;

    /**
     * Closes this iterator and linked iterators. 
     */
    public void close() {
	close(false);
    }

    /**
     * Closes this iterator and if not <tt>onlyThis</tt> also tail.
     */
    private void close(boolean onlyThis) {
	this.next = null;

	// Close this unless already closed
	if (!this.closed) {
	    this.closed = true;
	    try {
		closeHead();
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}

	// Close tail if requested there is one
	if (!onlyThis && this.tail != null) {
	    this.tail.close();
	}
    }

    /**
     * Closes this iterator and linked iterators. 
     */
    protected void finalize() {
	close();
    }

    /**
     * Unsupported operation.
     */
    public void remove() throws UnsupportedOperationException,
				IllegalStateException {
	throw new UnsupportedOperationException();
    }
}
