package fi.helsinki.w3m.metadata.item;

import com.hp.hpl.mesa.rdf.jena.model.* ;
import com.hp.hpl.mesa.rdf.jena.common.* ;
import com.hp.hpl.jena.util.* ;

import java.util.* ;

import fi.helsinki.w3m.metadata.*;
import fi.helsinki.w3m.metadata.ontology.*;

/*
 * $Revision: 1.36 $
 * $Date: 2002/05/17 12:31:21 $
 */
 
/**
*A class for representing the semantics of images.
*/
public class Image extends ConcreteItem {

	/* Override the OntologyResource property map. */
	protected static Hashtable predefinedPropertyHash;

	private boolean hasImageFile;
	
	static {
		/* Set URIs of property predicates that are predefined for this class by extending the OntologyResource property map. */
		Image.predefinedPropertyHash = new Hashtable(OntologyResource.predefinedPropertyHash);
		Image.predefinedPropertyHash.put(MuseumModel.DC_DESCRIPTION, MuseumModel.DC_DESCRIPTION);		
		Image.predefinedPropertyHash.put(MuseumModel.DC_SUBJECT, MuseumModel.DC_SUBJECT);		
	}

	/**
	*A constructor for the Image-object
	*
	*@param another Another Image-object to clone.
	*/
    public Image(Image another) {
	super(another);
	this.hasImageFile = false;
    }
	
	/**
	*A constructor for the Image-object
	*
	*@param model The parent model for this image.
	*@param id An Id-object for this image.
	*/
    public Image(MuseumModel model, Id id) {
	super(model, id);
	this.hasImageFile = false;	
    }

    /**
    *Properties that are predefined as methods in this class.
 	*
    *@return A Map-object holding the predefined properties of this ontology resource.
    */
    public Hashtable getPredefinedProperties() {
	return Image.predefinedPropertyHash;
    }

    /**
     *Returns the string description as definded by the dublic core dc:Description-tag.
     *
     *@return A string description of the image.
     */
    public String getDescription() {
	try {
	 return getNamedProperty(model.DC_DESCRIPTION, model.getBaseModel()).getValue().getValue().toString();
	} catch (Exception e) {
		return null;
	}
    }

    /**
     *Returns the string description as definded by the dublic core dc:Subject-tag. 
	 *Will probebly not be implemented now...
     *
     *@return A string subject of the image.
     */
    public String getSubject() {
	try {
	 return getNamedProperty(model.DC_SUBJECT, model.getBaseModel()).getValue().getValue().toString();
	} catch (Exception e) {
		return null;
	}
    }
	
	/**
	*Returns the label of this image.
	*
	*@return A string holding the label of the image.
	*/
	public String getLabel() {
		String subject = this.getSubject();
		if (subject == null) {
			return this.getShortCode();
		} else {
			return MuseumModel.decodeURI(subject);
		}		
	}
	
	/**
	*Returns the short code of this image.
	*
	*@return A string holding the short code of the image.
	*/
	protected String getShortCode() {
		String id = this.getId().toString();
		return id.substring(id.lastIndexOf(model.IMAGE_PREFIX) + model.IMAGE_PREFIX.length());
	}
	
	/**
	*Returns the ImageFile-objects associated with this image.
	*
	*@return A ResourceCollection-object holding the ImageFile-objects associated with this image. This should really be improved.
	*/
	public ImageFile getImageFile(final int typeCode) {
	try {
		PropertyImpl pro = new PropertyImpl(MuseumModel.IMFILE_NS + ImageFile.ID);		
	    SelectorImpl selector = new SelectorImpl(null, pro, this.getShortCode().toString() + "_" + typeCode);
		StmtIterator stmtiter = model.getImagefileModel().listStatements(selector);
		if (stmtiter.hasNext()) {
			Statement stmt = stmtiter.next();
			return new ImageFile(model, this, new SimpleId(stmt.getSubject().toString()), typeCode);
		} else {		
			this.hasImageFile = this.addFileStatements(typeCode);
			/* Avoid an infinite loop if addFileStatements() fails. */
			if (this.hasImageFile) {
				return this.getImageFile(typeCode);
			}
		}
	} catch (Exception e) {	
		e.printStackTrace();
	}		
	return null;
}

    /**
     *Returns the instances linked to this image and directly to the specified ontology class.
     *
	 *@param oclass The OntologyClass to which the returned instaces must belong.
     *@return A ResourceCollection-object holding the Image-objects for the images linked to this instance.
     */
	public ResourceCollection getInstancesForClass(final OntologyClass oclass) {
		StatementResourceCollection rv;
		try {
			ResourceImpl res = new ResourceImpl(this.getId().toString());		
			PropertyImpl pro = new PropertyImpl(oclass.getId().toString());
		    SelectorImpl selector = new SelectorImpl(res, pro, (String) null);
			ResourceProvider rp = new ResourceProvider() {
				public OntologyResource getResource( Object o ) {
					Statement stmt = (Statement) o;
					String object = stmt.getObject().toString();				
				    return new Instance(model, oclass, new SimpleId(object));
				}
	    	};	
			rv = new StatementResourceCollection(model.getBaseModel(), selector, rp);
		    return rv;
		} catch (RDFException rdfe) {
			rdfe.printStackTrace();
			return null;
		}
	}

/**
*Adds a base ImageFile of this Image to the Imagefile ontology. This should really be improved.
*
*@return True if the operation succeeded.
*/
private boolean addFileStatements(int typeCode) {
	/* Don't add statements if they are not needed. */
	try {
		ResourceImpl subjectres = new ResourceImpl((this.getShortCode() + "_" + typeCode), model.getImagefileModel());
		PropertyImpl idproperty = new PropertyImpl(model.IMFILE_NS + ImageFile.ID, model.getImagefileModel());
		LiteralImpl idliteral = new LiteralImpl(this.getShortCode() + "_" + typeCode);
		StatementImpl idstatement = new StatementImpl(subjectres, idproperty, idliteral);			
		model.getImagefileModel().add(idstatement);
		return true;			
  	} catch (Exception e) {		
		System.out.println("Error in Image.addFileStatements() while creating statements: "+e.toString());
 	}	
	return false;
}

}
