package fi.helsinki.w3m.metadata.item;

import java.util.*;
import java.net.*;

import com.hp.hpl.mesa.rdf.jena.model.* ;
import com.hp.hpl.mesa.rdf.jena.common.* ;
import com.hp.hpl.jena.util.* ;

import java.awt.Dimension;

import fi.helsinki.w3m.metadata.*;
import fi.helsinki.w3m.metadata.ontology.*;

/*
 * $Revision: 1.29 $
 * $Date: 2002/05/17 12:31:21 $
 */
 
public class ImageFile extends Image {

    /**
     *The Image which this class belongs to.
     */
    private Image parentImage;
	
	/**
	*The typecode of this imagefile.
	*/
	private int typeCode;
	
    /* Some magic strings specific to the Imagefile ontology */
    public final static String ID = "image_id";
    public final static String HEIGHT = "height";
    public final static String WIDTH = "width";
    public final static String COLOUR = "colour";
    public final static String FILENAME = "filename";	
    public final static String BASE_URL = "baseurl";
    public final static String FORMAT = "format";		
	public final static int ORIGINAL = 0;		
	public final static int THUMBNAIL = 1;	

    /* Override the OntologyResource property map. */
    protected static Hashtable predefinedPropertyHash;
	
    static {
	/* Set URIs of property predicates that are predefined for this class by extending the OntologyResource property map. */
	ImageFile.predefinedPropertyHash = new Hashtable(OntologyResource.predefinedPropertyHash);
	ImageFile.predefinedPropertyHash.put(ImageFile.ID, ImageFile.ID);		
	ImageFile.predefinedPropertyHash.put(ImageFile.HEIGHT, ImageFile.HEIGHT);		
	ImageFile.predefinedPropertyHash.put(ImageFile.WIDTH, ImageFile.WIDTH);		
	ImageFile.predefinedPropertyHash.put(ImageFile.COLOUR, ImageFile.COLOUR);		
	ImageFile.predefinedPropertyHash.put(ImageFile.FILENAME, ImageFile.FILENAME);		
	ImageFile.predefinedPropertyHash.put(ImageFile.BASE_URL, ImageFile.BASE_URL);		
	ImageFile.predefinedPropertyHash.put(ImageFile.FORMAT, ImageFile.FORMAT);		
    }
		
    /**
     *Constructor for the class. Makes copy of another ImageFile.
     *
     *@param another The ImageFile to be copied.
     */
    public ImageFile(ImageFile another) {
	super(another);
    }
	
    /**
     *Another constructor for this class.
     * 
     *@param model The model to which this class belongs to.
     *@param	parent	The Image to which this class belongs to.
     *@param	id	The identifier of this class.
     */
    public ImageFile(MuseumModel model, Image parent, Id id, int typeCode) {

	super(model, id);
	this.parentImage = parent;	
	this.typeCode = typeCode;	
    }
	
    /**
     *Properties that are predefined as methods in this class.
     *
     *@return A Map-object holding the predefined properties of this ontology resource.
     */
    public Hashtable getPredefinedProperties() {
	return ImageFile.predefinedPropertyHash;
    }

    /**
     *Returns the colour information of the imagefile.
     *
     *@return A String holding the colour information of the imagefile.
     */	
    public String getColour(){
	try {
	    return getNamedProperty(model.IMFILE_NS + this.COLOUR,model.getImagefileModel()).getValue().getValue().toString();
	} catch (Exception e) {
	    return "";
	}
    }
	
    /**
     *Returns the filename of the imagefile.
     *
     *@return A String holding the filename of the imagefile.
     */	
    public String getFileName(){
	try {
	    return getNamedProperty(model.IMFILE_NS + this.FILENAME, model.getImagefileModel()).getValue().getValue().toString();
	} catch (Exception e) {
	    return parentImage.getShortCode() + ".jpg";
	}
    }
	
    /**
     *Returns the base URL of the imagefile.
     *
     *@return A String holding the base URL of the imagefile.
     */	
    public String getBaseURL(){
	try {
	    String urlProperty = getNamedProperty(model.IMFILE_NS + this.BASE_URL, model.getImagefileModel()).getValue().getValue().toString();
	    return urlProperty;
	} catch (Exception e) {		
	    return model.getBaseURL().toString() + MuseumModel.IMAGE_PATH + this.typeCode + "/";
	}
    }

    /**
     *Returns the URL of the imagefile.
     *
     *@return An URL-object holding the URL of the imagefile.
     */	
	public URL getURL() {
		try {
			return new URL(getBaseURL() + getFileName());
		} catch (Exception e) {
		    return null;
		}			
	}

    /**
     *Returns the image format information of the imagefile.
     *
     *@return A String holding the image format information of the imagefile.
     */		
    public String getFormat(){
	try {
	    return getNamedProperty(model.IMFILE_NS + this.FORMAT, model.getImagefileModel()).getValue().getValue().toString();
	} catch (Exception e) {
	    return "";
	}
    }

    /**
     *Returns the height (in pixels) of the imagefile.
     *
     *@return A String holding the height (in pixels) of the imagefile.
     */		
    public int getHeight() {
	int height = -1;
	try {
	    height =
		getNamedProperty(model.IMFILE_NS + this.HEIGHT, model.getImagefileModel())
		.getValue()
		.getInt();
	} catch (Exception e) {}

	if (height < 0) {
	    height = getSize().height;
	}
	return height;
    }

    /**
     *Returns the id of the image this imagefile belongs to.
     *
     *@return A String holding the id of the image this imagefile belongs to.
     */		
    public Id getImageId() {
	return parentImage.getId();
    }

    /**
     *Retrieves the size of this ImageFile and saves it.
	 *
	 *@return A Dimension-object holding the height and width of this ImageFile.
     */
    protected Dimension getSize() {
	Dimension size = new Dimension(0,0);
 
        // Couldn't use Swing because that needs an X-server - doh
        ImageInfo ii = new ImageInfo();
	try {
          ii.setInput(getURL().openStream());
          if (ii.check())
  	  size = new Dimension(ii.getWidth(), ii.getHeight());
        } catch (Exception e) {
        }

	// Save width and height
	try {
		ResourceImpl subjectres = new ResourceImpl((this.getLabel() + "_"+this.typeCode), model.getImagefileModel());
		PropertyImpl heightproperty = new PropertyImpl(model.IMFILE_NS + ImageFile.HEIGHT, model.getImagefileModel());
		PropertyImpl widthproperty = new PropertyImpl(model.IMFILE_NS + ImageFile.WIDTH, model.getImagefileModel());		
		LiteralImpl heightliteral = new LiteralImpl(""+size.height);
		LiteralImpl widthliteral = new LiteralImpl(""+size.width);		
		StatementImpl heightstatement = new StatementImpl(subjectres, heightproperty, heightliteral);			
		StatementImpl widthstatement = new StatementImpl(subjectres, widthproperty, widthliteral);					
		model.getImagefileModel().add(heightstatement);
		model.getImagefileModel().add(widthstatement);		
   	} catch (Exception Ex) {		
		System.out.println("Error in ImageFile.getSize() while adding statements: "+Ex.toString());
   	}		
	return size;
    }
	
    /**
     *Returns the width (in pixels) of the imagefile.
     *
     *@return A String holding the width (in pixels) of the imagefile.
     */		
    public int getWidth() {
	int width = -1;
	try {
	    width =
		getNamedProperty(model.IMFILE_NS + this.WIDTH, model.getImagefileModel())
		.getValue()
		.getInt();
	} catch (Exception e) {}

	if (width < 0) {
	    width = getSize().width;
	}
	return width;
    }
	
}
