package fi.helsinki.w3m.metadata.ontology;

import java.util.*;
import java.net.*;
import java.lang.reflect.*;

import com.hp.hpl.mesa.rdf.jena.model.* ;
import com.hp.hpl.mesa.rdf.jena.common.* ;
import com.hp.hpl.jena.util.* ;

import fi.helsinki.w3m.metadata.*;
import fi.helsinki.w3m.metadata.item.*;

/*
 * $Revision: 1.49 $
 * $Date: 2002/05/17 12:31:22 $
 */

/**
*Represents a node in the model ontology network.
*/
public class OntologyClass extends OntologyElement {
    public static final int ALL = -1;
    public static final int DIRECT = 0;


   /**
    *This is used to avoid the need to recount the image count for an ontology class.
	*/
	private static Hashtable imageCountHash;

	static {
	    /* Create the hashtable once for this class. */	
	    imageCountHash = new Hashtable();
	}

    /**
     *Constructor for the OntologyClass class.
     *
     *@param another An OntologyClass-object to clone.
     */
	public OntologyClass(OntologyClass another) {
	super(another);
	this.label = this.getId().toString().substring(this.getId().toString().lastIndexOf("#") + 1);	
    }

    /**
     *Constructor for the OntologyClass class.
     *
     *@param id An Id-object holding the literal name of this ontology class.
     *@param model A MuseumModel-object holding the model this object is a part of.
     */
    public OntologyClass(MuseumModel model, Id id) {
	super(model, id);
	this.label = this.getId().toString().substring(this.getId().toString().lastIndexOf("#") + 1);	
    }

    /**
     *Returns the children of this ontology class and its subtree.
     *
     *@param depth An int value holding the depth of the subtree to traverse in order to find children of this ontology class. Specify DIRECT not search any subtree and ALL to search the entire subtree.
     *@return A ResourceCollection-object holding the OntologyClass-objects of the children of this ontology-class and its depth-sized subtree.
     */
    public ResourceCollection getChildren(int depth) {
	StatementResourceCollection rv = null;
	try {
	    ResourceImpl res = new ResourceImpl(this.getId().toString());
	    PropertyImpl pro = new PropertyImpl(model.SUBCLASSOF_PROPERTY_URI);
	    SelectorImpl selector = new SelectorImpl(null, pro, res);
	    ResourceProvider rp = new ResourceProvider() {
		    public OntologyResource getResource( Object o ) {
			Statement stmt = (Statement) o;
			return new OntologyClass( model, new SimpleId(stmt.getSubject().toString()) );
		    }
		};				
	    rv = new StatementResourceCollection(model.getBaseModel(), selector, rp);
		/* Recursively get depth-1-sized subtree. Ignore depths of 0. */  
		if (depth != DIRECT) {
	    	if (depth > 0) depth--;
		    Iterator iter = rv.iterator();
		    while (iter.hasNext()) {
				OntologyClass aChild = (OntologyClass)iter.next();
				if (aChild != null) {
		    		ResourceCollection aChildColl = aChild.getChildren(depth); 
		    		rv.addAll(aChildColl);
				}
	    	}
	   	}
	} catch (Exception e) {
		e.printStackTrace();
	}
	return rv;
	}

    /**
     *Returns the parents of this ontology class and its supertree.
     *
     *@param depth An int value holding the depth of the supertree to traverse in order to find parents of this ontology class. Specify DIRECT not search any supertree and ALL to search the entire supertree.
     *@return A ResourceCollection-object holding the OntologyClass-objects of the parents of this ontology-class and its depth-sized supertree.
     */
    public ResourceCollection getParents(int depth) {
	StatementResourceCollection rv;
	try {
	    ResourceImpl res = new ResourceImpl(this.getId().toString());		
	    PropertyImpl pro = new PropertyImpl(model.SUBCLASSOF_PROPERTY_URI);
	    SelectorImpl selector=new SelectorImpl(res, pro, (String) null);
	    ResourceProvider rp = new ResourceProvider() {
		    public OntologyResource getResource( Object o ) {
			/* TODO: filter out http://www.w3.org/TR/1999/PR-rdf-schema-19990303#Resource more cleanly */			
		    Statement stmt = (Statement) o;
		    if ( stmt.getObject().toString().equals(model.RESOURCE_CLASS_URI) )
				return null;
		    else 
				return new OntologyClass( model, new SimpleId(stmt.getObject().toString()) );
		    }
		};	
	    rv = new StatementResourceCollection(model.getBaseModel(), selector, rp);
	} catch (RDFException rdfe) {
	    rdfe.printStackTrace();
	   	return null;
	}	
		
	/* Recursively get depth-1-sized subtree. Ignore depths of 0. */  
	if (depth != DIRECT) {
	    if (depth > 0) depth--;
	    Iterator iter = rv.iterator();
	    while (iter.hasNext()) {
		OntologyClass aParent = (OntologyClass)iter.next();
		if (aParent != null) {
		    ResourceCollection aParentColl = aParent.getParents(depth); 
		    rv.addAll(aParentColl);
		}
	    }
	}
	return rv;
    }

    /**
     *Returns the instances of this ontology class.
     *
     *@return A ResourceCollection-object holding the Instance-objects of the instances of this ontology-class and its depth-sized subtree.
     */
    public ResourceCollection getInstances(int depth) {
		StatementResourceCollection instances;	
		try {
			ResourceImpl res = new ResourceImpl(this.getId().toString());		
			PropertyImpl pro = new PropertyImpl(model.TYPE_PROPERTY_URI);
			SelectorImpl selector=new SelectorImpl(null, pro, res);
			final OntologyClass PARENT_CLASS = this;
			ResourceProvider rp = new ResourceProvider() {
				public OntologyResource getResource( Object o ) {
					Statement stmt = (Statement) o;
			    	return new Instance( model, PARENT_CLASS, new SimpleId(stmt.getSubject().toString()) );
				}
		    };
			instances = new StatementResourceCollection(model.getBaseModel(), selector, rp);

		} catch (Exception rdfe) {
			rdfe.printStackTrace();
			return null;
		}
		
		/* Recursively get depth-1-sized subtree. Ignore depths of 0. */  		
		if (depth != DIRECT) {
		    if (depth > 0) depth--;
			/* Get all instances of all the children unless they are insane, in which case this is the end. */
			ResourceCollection aChildColl = getChildren(depth);	
			Iterator iter = aChildColl.iterator();
		while (iter.hasNext()) {
			OntologyClass aChild = (OntologyClass)iter.next();
				instances.addAll(aChild.getInstances(DIRECT));
			}
		/* Avoid duplicates. */
		LinkedList wrapper = new LinkedList();
		wrapper.add(instances);					
		return intersect(wrapper); 
		}
		return instances;
	}
	
    /**
     *Returns the images linked to this ontology class through its instances.
     *
	 *@see Instance#getImages()
     *@param depth An int value holding the depth of the subtree to traverse in order to find images of this ontology class. Specify DIRECT not search any subtree and ALL to search the entire subtree.
     *@return A ResourceCollection-object holding the images linked to this ontology class through its instances.
     */
	public ResourceCollection getImages(int depth) {
		SimpleResourceCollection baseClassImages = new SimpleResourceCollection(new LinkedList());
		try {		
			ResourceCollection instances = this.getInstances(depth);
			Iterator iter = instances.iterator();			
			while (iter.hasNext()) {
				Instance anInstance = (Instance)iter.next();
				baseClassImages.addAll(anInstance.getImages());
			}				
			/* Avoid duplicates. */
			LinkedList wrapper = new LinkedList();
			wrapper.add(baseClassImages);					
			return intersect(wrapper); 
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
     /**
     *Intersects the images of this object with an existing intersection
     *
     *@param intersectImages A Hashtable-object holding the OntologyClass-objects of an intersection.
     *@param depth An int value holding the depth of the intersection.
     *@return A ResourceCollection-object holding the intersected Image-objects.
     */
    public ResourceCollection getIntersectImages (Hashtable intersectImages, int depth) {
	ResourceCollection baseClassImages = this.getImages(depth);
	if (intersectImages == null) return baseClassImages;
	SimpleResourceCollection intersection = new SimpleResourceCollection(new LinkedList());
	
	try {		
		/* FIXME: Use intersect() in OntologyResource */
		Iterator i = baseClassImages.iterator();
		while (i.hasNext()) {
			Image img = (Image)i.next();
		    if (intersectImages.containsKey(img)) {
			intersection.add(img);
		    }
		}	
	} catch (Exception e) {
		e.printStackTrace();
	}
	return intersection;
    }
	
    /**
     * Returns the number of images linked to an ontology class.
	 *
	 *@return A long value holding the number of images linked to this ontology class.
     */
    public long getImageCount() {
		if (imageCountHash.containsKey(this.getId())) {			
			return ((Long)imageCountHash.get(this.getId())).longValue();
		} 
		try {
			long rv = 0;								
			ResourceCollection instances = this.getImages(ALL);
			Iterator iter = instances.iterator();			
			while (iter.hasNext()) {
				iter.next();
				rv++;
			}				
			/* Save the count in the count hash */
			imageCountHash.put(this.getId(), new Long(rv));			
			return rv;
		} catch (Exception e) {
			e.printStackTrace();
			return -1;
		}
	}
}
