package fi.helsinki.w3m.metadata.ontology;

import fi.helsinki.w3m.metadata.*;
import fi.helsinki.w3m.metadata.item.*;

import java.util.*;
import java.net.*;

import com.hp.hpl.mesa.rdf.jena.model.* ;
import com.hp.hpl.mesa.rdf.jena.common.* ;
import com.hp.hpl.jena.util.* ;

/*
 * $Revision: 1.62 $
 * $Date: 2002/05/23 15:33:17 $
 */

/**
*An abstract class implementing common features for all OntologyResource- classes.
*/
public abstract class OntologyResource implements Comparable {
    protected final Id id;
    protected String label; 
    protected final MuseumModel model;
	protected static Hashtable predefinedPropertyHash;

	static {
		/* Set URIs of property predicates that are predefined for this class */
		predefinedPropertyHash = new Hashtable();
		predefinedPropertyHash.put(MuseumModel.TYPE_PROPERTY_URI, MuseumModel.TYPE_PROPERTY_URI);
		predefinedPropertyHash.put(MuseumModel.LABEL_PROPERTY_URI, MuseumModel.LABEL_PROPERTY_URI);		
	}

	/**
	*A constructor for the OntologyResource class.
	*
	*@param another An OntologyResource object to clone.
	*/
    public OntologyResource(OntologyResource another) {
	this(another.model, another.id);
    }

	/**
	*A constructor for the OntologyResource class.
	*
	*@param model The parent MudeumModel of this object.
	*@param id The identifier for this object.
	*/
    public OntologyResource(MuseumModel model, Id id) {
	if (id != null) {
		this.id = id;
	} else {
		throw new IllegalArgumentException("Id is null in OntologyResource constructor.");
	}
	this.model = model;
	this.label = getNameForURI(this.id.getId());
    }

	/**
	*A constructor for the OntologyResource class.
	*
	*@param model The parent MudeumModel of this resource.
	*@param label A string holding the specified label for this resource.
	*@param id The identifier for this resource.
	*/
    public OntologyResource(MuseumModel model,String label, Id id) {
	if (id != null) {
		this.id = id;
	} else {
		throw new IllegalArgumentException("Id is null in OntologyResource constructor.");
	}
	this.model = model;
	this.label = label;
    }

	/**
	*This method implements the Comparable-interface.
	*/
    public int compareTo(Object anotherOntologyResource) throws ClassCastException{
	OntologyResource another = (OntologyResource)anotherOntologyResource;
	int rv = 0;

	if ( !this.isComparable() || !another.isComparable() )
	    throw new ClassCastException();
	
	// First try the short keys and...
	Comparable thisKey = this.getLocalKey();
	Comparable anotherKey = another.getLocalKey();

	// ...only if they match...
	if ( thisKey != null && anotherKey != null ) {
	    rv = thisKey.compareTo( anotherKey ); 
	}

	// ...compare the actual keys.
	if ( rv == 0 ) {
	    thisKey = this.getKey();
	    anotherKey = another.getKey();
	    rv = thisKey.compareTo( anotherKey );
	}
	return rv;
    }
	
	/**
	*Indicates whether some other OntologyResource is "equal to" this one.
	*
	*@see String#equals(Object) equals
	*/
    public boolean equals(Object another) {
	return this.compareTo( another ) == 0;
    }

    protected static String getNameForURI(String uri) {
	String rv = null;
	if (uri != null) {
	    int i = uri.indexOf(MuseumModel.NAME_PREFIX);
	    i++;
	    if (i > 0 && i < uri.length()) {
		rv = java.net.URLDecoder.decode(uri.substring(i));
	    }
	}
	return rv;
    }

	/**
	*Returns the intersection of the specified collections.
	*
	*@param collections A collection of collections to intersect.
	*@return A Hashtable-collection holding the intersection of the specified collections.
	*/
    public static Hashtable getIntersectHash(Collection collections) throws IllegalArgumentException{
	Hashtable intersect = new Hashtable();
	Enumeration enum = null;
	Iterator i = collections.iterator();
	Iterator j = null;
	int n = 0;

	// Create hash.
	while (i.hasNext()) {
	    Object tmp = i.next();
	    boolean singleCollection = false;

	    // Collection of collections...
	    if (tmp instanceof Collection) {
		j = ((Collection)tmp).iterator();
	    }

	    // ...or single collection of OntologyResource's
	    else if (tmp instanceof OntologyResource) {
		singleCollection = true;
		j = collections.iterator();
	    }
	    else {
		throw new IllegalArgumentException();
	    }

	    while (j.hasNext()) {
		OntologyResource or = (OntologyResource)j.next();
		if (n == 0) {
		    intersect.put(or, new Integer(1));
		}
		else {
		    Object key = intersect.get(or);
		    if (key != null) {
			int count = ((Integer)key).intValue();
			// This element doesn't belong in some previous collection.
			if (count < n) {
			    intersect.remove(or);
			}
			// Each item in collection j will be added only once.
			else if (count == n) {
			    intersect.put(or, new Integer(count+1));
			}
		    }
		}
	    }
	    n++;
	    if (singleCollection) break;
	}
	// Remove values that are not incommon with all the collections.
	enum = intersect.keys();
	while (enum.hasMoreElements()) {
	    Object key = enum.nextElement();
	    Object value = intersect.get(key);
	    if (value != null) {
		int count = ((Integer)value).intValue();
		if (count < n) {
		    intersect.remove(key);
		} else if (count == n) {
		    intersect.put(key, key);
		} else {
		    System.out.println("Detected a bug in the getIntersectHash-method!");
		}
	    }
	}
	return intersect;
    }

	/**
	*Returns the intersection of the specified collections.
	*
	*@param collections A collection of collections to intersect.
	*@return A ResourceCollection-object holding the intersection of the specified collections.
	*/
	public static ResourceCollection intersect(Collection collections) {
	return intersect(getIntersectHash(collections));
    }

	/**
	*Returns a ResourceCollection of the specified Hashtable-intersection.
	*
	*@param intersect A Hashtable of an intersection.
	*@return A ResourceCollection-object holding the intersection.
	*/
    public static ResourceCollection intersect(Hashtable intersect) {
	return new SimpleResourceCollection(intersect.values());
    }

	/**
	*Returns the comparable key of this ontology resource in its namespace.
	*
	*@return A Comparable-object holding the key of this ontology resource in its namespace.
	*/
    protected Comparable getLocalKey() { return this.label; }

	/**
	*Returns the comparable key of this ontology resource.
	*
	*@return A Comparable-object holding the key of this ontology resource.
	*/
    protected Comparable getKey() { return this.id; }

	/**
	 *Returns a java.net.URL-object holding the URL for this OntologyResource. May not work on all resources.
	 *
 	 *@return A java.net.URL-object holding the URL for this OntologyResource.
	 */
    public URL getURL() throws MalformedURLException { return new URL(this.id.getId()); }

    protected Resource getResource() {
	if (this.id instanceof CompositeId) return null;
	else return new ResourceImpl(this.getId().toString());
    }

	/**
	*Returns the identifier for this resource.
	*
	*@return An Id-object holding the identifier for this resource.
	*/
    public Id getId() { return this.id; }

    /**
    *Properties that are predefined as methods in this class.
 	*
    *@return A Map-object holding the predefined properties of this ontology resource.
    */
    public Hashtable getPredefinedProperties() {
	return OntologyResource.predefinedPropertyHash;
    }

   /**
    *Returns the (human readable) label for this resource.
 	*
    *@return A string holding the (human readable) label for this resource.
    */
	public String getLabel() { 
		return model.decodeURI(this.label); 
	}

   /**
    *Returns the short code for this resource.
 	*
    *@return A string holding the short code for this resource.
    */
	protected String getShortCode() {
		return this.getLabel();
	}

	/**
	*Returns the non-predefined (see getPredefinedProperties above) properties of this ontology resource.
	*
	*@return A Collection-object holding the non-predefined (see getPredefinedProperties above) properties of this ontology resource.
	*/
	public ResourceCollection getProperties() {
		try { 
			ResourceCollection rv = null;
			ResourceImpl res = new ResourceImpl(this.getId().toString());		
			SelectorImpl selector = new SelectorImpl( res, null, (String) null );
			final OntologyResource PARENT = this;
			ResourceProvider rp = new ResourceProvider() {
				public OntologyResource getResource( Object o ) {					
					Statement stmt = (Statement) o;
					if (! PARENT.getPredefinedProperties().containsKey(stmt.getPredicate().toString())) {
						SimpleId ownerId = (SimpleId)PARENT.getId();
						SimpleId propertyId = new SimpleId(stmt.getPredicate().toString());
						PropertyDefinition pd = new PropertyDefinition(PARENT.model, new CompositeId(ownerId, propertyId));
						Object value;
						String valueAsString = stmt.getObject().toString();
						if (valueAsString.startsWith(MuseumModel.BASE_NS+MuseumModel.INSTANCE_PREFIX)) {
							value = new Instance(PARENT.model, new SimpleId(valueAsString));
						/*} else if (valueAsString.startsWith(MuseumModel.BASE_NS+MuseumModel.IMAGE_PREFIX)) {
							value = new Image(PARENT.model, new SimpleId(valueAsString));													
						} else if (valueAsString.startsWith(MuseumModel.BASE_NS)) {
							value = new OntologyClass(PARENT.model, new SimpleId(valueAsString));*/
						} else {
							value = new String(valueAsString);
						}
						return new ResourceProperty(pd, PARENT, value, PARENT.model);
					} else return null;
				}
			   };
			rv = new StatementResourceCollection(model.getBaseModel(), selector, rp);
			return rv;
		 } catch (Exception e) {
		 	e.printStackTrace();
		 }
		 return null;
	} 

    /**
     *Sets the weight of this ontology resource to the specified value.
     *
     *@return A long-value holding the new weight for this ontology resource.
     */
    public long setWeight(long value) {
	try {	
		/* Get the weight as a string and cast it to a long value */	
		Statement weightStatement = getWeightValueStatement();
	    weightStatement.set(""+value); 		
	    return value;				
	} catch (Exception e) {
		/* Probable cause: no weight statements for this OntologyResource */
		if (addWeightStatements()) {
			return 1;
		} else {	
		    return 0;
		}
	}
    }
	
    /**
     *Increases the weight of this ontology resource by one (1).
     *
     *@return A long-value holding the new weight for this ontology resource.
     */
	public long addWeight() {
		return this.setWeight(this.getWeight() + 1);	
	}

    /**
     *Returns the weight for this ontology resource.
     *
     *@return A long-value holding the weight for this ontology resource.
     */
    public long getWeight() {
	try {	
		/* Get the weight as a string and cast it to a long value */
		Statement weightStatement = getWeightValueStatement();
	    String weightString = weightStatement.getObject().toString();
	    Long weightCast = new Long(weightString);		
	    long weight = weightCast.longValue();				
	    return weight;				
	} catch (Exception e) {
		    return 0;
	}
    }

    /**
     *Returns the statement holding the weight for this ontology resource.
     *
     *@return A Statement holding the weight for this ontology resource.
     */
	private Statement getWeightValueStatement() {
	try {
	    /* Get the weight-instance for this resource. */
		PropertyImpl pro = new PropertyImpl(model.WEIGHT_ID_PROPERTY_URI);		
	    SelectorImpl selector = new SelectorImpl(null, pro, this.getId().toString());			
	    StmtIterator stmtIter = model.getWeightModel().listStatements(selector);
	    Statement stmt = stmtIter.next();
	    String subject = stmt.getSubject().toString();
	    String weightInstanceID = subject;
        
	    /* Get the weight-count of the weight-instance found above. */
		ResourceImpl res2 = new	ResourceImpl(weightInstanceID);
		PropertyImpl pro2 = new PropertyImpl(model.WEIGHT_COUNT_PROPERTY_URI);
	    SelectorImpl weightselector = new SelectorImpl(res2, pro2, (String) null);
	    StmtIterator weightStmtIter = model.getWeightModel().listStatements(weightselector);

 	    return weightStmtIter.next();
	} catch (Exception e) {	
		return null;
	}		
    }

	/**
	*Add weight statements for this OntologyResource to the weightmodel. 
	*
	*@return A boolean value indicating the success of the operation.
	*/
	private boolean addWeightStatements () {
	    try {		
		/* Create the parts of the four statment triples */
		String idToWrite;		
		if (this.getId().toString().indexOf(model.NAME_PREFIX) > 0 || this.getId().toString().startsWith(model.IMAGE_PROTOCOL)) {
			idToWrite = this.getShortCode();
		} else {
			idToWrite = this.getId().toString();
		}
		ResourceImpl subject = new ResourceImpl(model.WEIGHT_NS + idToWrite + model.WEIGHT_SUFFIX, model.getWeightModel());
		ResourceImpl classresource = new ResourceImpl(model.RESOURCE_WEIGHT_URI, model.getWeightModel());
				
		PropertyImpl typeproperty = new PropertyImpl(model.TYPE_PROPERTY_URI, model.getWeightModel());
		PropertyImpl idproperty = new PropertyImpl(model.WEIGHT_ID_PROPERTY_URI, model.getWeightModel());
		PropertyImpl weightproperty = new PropertyImpl(model.WEIGHT_COUNT_PROPERTY_URI, model.getWeightModel());
		PropertyImpl labelproperty = new PropertyImpl(model.LABEL_PROPERTY_URI, model.getWeightModel());						
		
		/* We add weight the first time here, so a magic value of 1 is ok. */		
		LiteralImpl weightliteral = new LiteralImpl("1");
		LiteralImpl idliteral = new LiteralImpl(this.getId().toString());
		LiteralImpl labelliteral = new LiteralImpl(this.getId().toString() + model.WEIGHT_SUFFIX);
			
		/* Create the statements and add them to the model. They will be saved when the model is written to disk. */
		StatementImpl typestatement = new StatementImpl(subject, typeproperty, classresource);
		StatementImpl labelstatement = new StatementImpl(subject, idproperty, idliteral);
		StatementImpl idstatement = new StatementImpl(subject, labelproperty, labelliteral);
		StatementImpl weightstatement = new StatementImpl(subject, weightproperty, weightliteral);									
			
		model.getWeightModel().add(typestatement);
		model.getWeightModel().add(labelstatement);
		model.getWeightModel().add(idstatement);
		model.getWeightModel().add(weightstatement);
		
		return true;					
	    } catch (Exception e) {		
			e.printStackTrace();
			return false;
	    }		
	}
	
	/**
	*Returns the MuseumModel this resource belongs to.
	*
	*@return A MuseumModel-object holding the MuseumModel this resource belongs to.
	*/
	public MuseumModel getModel(){
		return this.model;
	}
	
	/**
	*Returns the specified non-predefined property of this ontology resource.
	*
	*@return A ResourceProperty-object holding the specified non-predefined property of this ontology resource.
	*/
	public ResourceProperty getNamedProperty(String value, Model jenaModel){
		try {
			ResourceCollection rv = null;
			ResourceImpl res = new ResourceImpl(this.getId().toString());
			PropertyImpl pro = new PropertyImpl(value);
		    SelectorImpl selector = new SelectorImpl(res, pro, (String)null);
			final OntologyResource PARENT = this;
			ResourceProvider rp = new ResourceProvider() {
				public OntologyResource getResource( Object o ) {					
					Statement stmt = (Statement) o;
                    SimpleId propertyId = new SimpleId(stmt.getPredicate().toString());
					SimpleId ownerId = (SimpleId)PARENT.getId();
					PropertyDefinition pd = new PropertyDefinition(PARENT.model, new CompositeId(ownerId, propertyId));
					Object value;
					String valueAsString = stmt.getObject().toString();
					if (valueAsString.startsWith(MuseumModel.BASE_NS+MuseumModel.INSTANCE_PREFIX)) {
						value = new Instance(PARENT.model, new SimpleId(valueAsString));
					/*} else if (valueAsString.startsWith(MuseumModel.BASE_NS+MuseumModel.IMAGE_PREFIX)) {
						value = new Image(PARENT.model, new SimpleId(valueAsString));													
					} else if (valueAsString.startsWith(MuseumModel.BASE_NS)) {
						value = new OntologyClass(PARENT.model, new SimpleId(valueAsString));*/
					} else {
						value = new String(valueAsString);
					}
					return new ResourceProperty(pd, PARENT, value, PARENT.model);
				}
			   };
			rv = new StatementResourceCollection(jenaModel, selector, rp);
			Iterator iteri=rv.iterator();
			return (ResourceProperty)iteri.next();	
														
		} catch (RDFException rdfe) {
			rdfe.printStackTrace();
			return null;
		}		
    }	
	
	/**
	*Implements the Comparable interface.
	*/
    public boolean isComparable() { return true; }

	/**
	*Make hashcodes of OntologyResources useful.
	*
	*@return An int value holding a hashcode for this OntologyResource.
	*/
	public int hashCode() {
		return this.id.getString().hashCode();
	}
	
	public String toString() {
		return this.getLabel();
	}	
}
