/**
 * University of Helsinki
 * WWW-Museum
 *
 * @author Pasi Lehtimki and Samppa Saarela
 * Profile contains information about current user's previous selectons and selected facets
 */
package fi.helsinki.w3m.profile;

import fi.helsinki.w3m.*;

import fi.helsinki.w3m.beans.*;

import fi.helsinki.w3m.metadata.ontology.OntologyClass;
import fi.helsinki.w3m.metadata.ontology.Instance;
import fi.helsinki.w3m.metadata.item.Image;
import fi.helsinki.w3m.metadata.SimpleId;
import fi.helsinki.w3m.metadata.MuseumModel;
import fi.helsinki.w3m.filter.Filter;
import fi.helsinki.w3m.filter.InterestingOntologyClass;

import java.util.*;


public class Profile{
	
    /**
     * Separates class ids in selection paths
     */
    public static final String ONTOLOGY_CLASS_SEPARATOR = ";";
    
    /**
     * The number of facets shown to the user in default
     */
    public static final int NO_OF_DEFAULT_FACETS = 3;

    /**
     * Ordering information of the facet
     */
    public static final int ORDER_BY_INSTANCES = 0;
    public static final int ORDER_BY_IMAGES = 1;

    //base URL of icon images
    private static final String ICON_URL = "http://www.cs.helsinki.fi/group/wwwmuseo/images/";
    
    //maximum size of image shown to the user
    private int maxImageWidth;
    private int maxImageHeight;
    
    //helper variable for creating FacetSelection ids
    private int facetId = 0;
    
    //helper for counting the intersect
    private Hashtable intersect;    
    
    //Vector of FacetSelection-objects. Containing all visible facets
    private Vector selectedFacets;
	
    //Vector contining Strings. Image IDs
    private Vector previousSelections;

    //???
    private Vector allFacetRoots;
    
    //Instances for using Filter and MuseumModel
    private Filter filter;
    private MuseumModel model;
	
    /**
     * Constructs new Profile
     */
    public Profile(){
	    //System.out.println("new profile");
	this.maxImageHeight = 300;
	this.maxImageWidth = 400;
	   
	String location = "/home/tkt_w3mu/wwwmuseo_group/wwwmuseo_webapp/WEB-INF/data/";
	filter = W3m.getFilter(location);
	model = W3m.getModel(location);
	selectedFacets = new Vector();
	previousSelections = new Vector();
	allFacetRoots = new Vector();
	initializeFacets();
    }

    
    /**
     * @return maximum width of image
     */
    public int getMaxImageWidth(){
	    return this.maxImageWidth;
    }
    
    /**
     * Sets maximum width of image
     */    
    public void setMaxImageWidth(int i){
	    this.maxImageWidth = i;
    }

    /**
     * @return maximum height of image
     */       
    public int getMaxImageHeight(){
	    return this.maxImageHeight;
    }
    
    /**
     * Sets maximum height of image
     */       
    public void setMaxImageHeight(int i){
	    this.maxImageHeight = i;
    }    
    
    /*
     * Initializes facets
     */
    private void initializeFacets() {
	Vector v = null;
	try{
		v = new Vector(filter.getFacetRoots(this));
		//System.out.println("got roots:" + v.size());
	}catch(Exception e){
		e.printStackTrace();
	}

	//System.out.println("initializing facets");

	int fscount = 0;
	for (int i = 0; i < v.size() && fscount < NO_OF_DEFAULT_FACETS; ++i) {
	    FacetSelection f = new FacetSelection( getNewFacetId(), (OntologyClass)v.elementAt(i) );
	    if (f.getFacetRoot().getImageCount() > 0) {
		this.selectedFacets.add(f);
		fscount++;
	    }
	}
    }

    /**
     * Returns all selected classes
     * @return Vector containing OntologyClass objects.
     */

    public Vector getActiveClasses() {
	Vector aclasses = new Vector();

	//System.out.println("return state of hibrowse");
	for ( int i=0; i < this.selectedFacets.size(); i++ ) {
	    if(((FacetSelection)this.selectedFacets.get(i)).getActiveNode() != null)
		    aclasses.add( ((FacetSelection)this.selectedFacets.get(i)).getActiveNode() );
	}
	return aclasses;
    }

    
    /*
     * Used to create ids for FacetSelections
     */
    protected int getNewFacetId() {
	return this.facetId++;
    }

    /**
     * Used by UI to get list of facets that can be added
     * @return Vector containing AddableFacetBean objects
     */
    public Vector getAddableFacetBeans(){
	    Vector ret = new Vector();
	    
	    try{
		Vector allFacets = new Vector(filter.getFacetRoots(this));

	    
	    	for (int i = 0; i < allFacets.size(); ++i) {
			OntologyClass facet = (OntologyClass)allFacets.elementAt(i);
			AddableFacetBean afb = new AddableFacetBean();
			afb.setId(facet.getId().getString());
		    	afb.setName(facet.getLabel());
			afb.setImageCount((int)facet.getImageCount());
			ret.add(afb);
		}
	    }catch(Exception e){e.printStackTrace();}

	    
	    return ret;
    }
   
    /**
    * @return Vector containing FacetBean objects
    */
    public Vector getFacetBeans(){

	    Vector v = getActiveClasses();
	     
	    if (v.size() > 0) {
		    Vector images = new	Vector();	    
		    for (int i=0; i < v.size(); ++i) {
			    images.add(((OntologyClass)v.get(i)).getImages(OntologyClass.ALL));
		    }
		    //System.out.println("update intersect");
		    this.intersect = OntologyClass.getIntersectHash(images);
	    }else{
		    //System.out.println("else");
		    this.intersect = OntologyClass.getIntersectHash(model.getAllImages());
	    
	    }
	    
	    Vector facetBeans = new Vector();

	    for(int i = 0; i < this.selectedFacets.size(); ++i){
		    FacetBean fb = new FacetBean();
		    FacetSelection facetSelection = (FacetSelection)this.selectedFacets.elementAt(i);
		    fb.setFacetId( facetSelection.getId() );
		    fb.setFacetName(facetSelection.getName());
		    
		    
		    StringTokenizer st = new StringTokenizer(facetSelection.getSelectionPath(), Profile.ONTOLOGY_CLASS_SEPARATOR);
		    
		    String facetPathString = "";
		    String tmp = "";
		    while (st.hasMoreTokens()) {
			    String parentId = st.nextToken();
			    OntologyClass parent = model.getOntologyClass(new SimpleId(parentId));
			    
			    tmp = tmp+"<a href=\"/wwwmuseo/servlets/ActionServlet?"+
			    	"paramNames=selectionPath,facet,class"+
				"&action=selectClass"+
				"&selectionPath="+java.net.URLEncoder.encode(facetPathString)+
				"&facet="+facetSelection.getId()+
				"&target=hiBrowse"+
				"&class="+
				java.net.URLEncoder.encode(parent.getId().getString())+
				"&debug=jotain\" class=\"facetRoot\">"+
				parent.getLabel()+"</a><b>/</b>";
				
			facetPathString += Profile.ONTOLOGY_CLASS_SEPARATOR;
			facetPathString += parentId;
		    }
		    
		    if(facetSelection.getActiveNode() != null){
			    OntologyClass parent = model.getOntologyClass(facetSelection.getActiveNode().getId());
		    
		    	tmp = tmp+"<a href=\"/wwwmuseo/servlets/ActionServlet?"+
			    	"paramNames=selectionPath,facet,class"+
				"&action=selectClass"+
				"&selectionPath="+java.net.URLEncoder.encode(facetPathString)+
				"&facet="+facetSelection.getId()+
				"&target=hiBrowse"+
				"&class="+
				java.net.URLEncoder.encode(parent.getId().getString())+
				"&debug=jotain\" class=\"facetRoot\">"+
				parent.getLabel()+"</a>";
		    }
		    
		    fb.setHeaderNodes(tmp);
		    
		    Vector visibleNodes = new Vector();
		    OntologyClass node = facetSelection.getActiveNode();
		    if(node == null){
			    node = facetSelection.getFacetRoot();
		    }

		    // If facet is open...
		    if (facetSelection.isOpen(node.getId().getString())) {
			    addNodeBean(visibleNodes, facetSelection, facetSelection.getSelectionPath(), node, false, 1, true);
		    }else{
			    addNodeBean(visibleNodes, facetSelection, facetSelection.getSelectionPath(), node, false, 1, false); 
		    }
		    
		    
		    fb.setVisibleNodes(visibleNodes);
		    
		    facetBeans.add(fb);
	    }
	    return facetBeans;
    }
    
    
    /*
     * Helper method for creatinf facet's nodes that should be shown to the user.
     * This method is used recusively.
     */
    private void addNodeBean(Vector visibleNodes, FacetSelection facet, String selectionPath, OntologyClass node, boolean onlyContent, int depth, boolean rootOpen) {
	 //System.out.println("Starting addNodeBean(). Depth: "+depth);
	 Vector children = null;
	 Vector pics = null;
	 Vector picsAll = null;
	 
	 //System.out.println("active intrsect:" +this.intersect);

	//System.out.println("***");
	
	// Childrens of node
	children = new Vector(node.getChildren( OntologyClass.DIRECT ));
	
	//System.out.println("***");
	// Pictures that are direct instances of OntologyClass node
	pics = new Vector(node.getIntersectImages( this.intersect, OntologyClass.DIRECT ));
	//System.out.println("***");
	// All pictures that are  instances of OntologyClass node
	picsAll = new Vector(node.getIntersectImages(this.intersect, OntologyClass.ALL));


//System.out.println("pics size: "+pics.size());
//System.out.println("picsAll size: "+picsAll.size());	
	
	
	//System.out.println("***");

	boolean isOpen = facet.isOpen(node.getId().getString());

	boolean hasContent = children.size() > 0 || picsAll.size() > 0;

	String expanderOrCollapserHTML = "";
	String id = node.getId().getString();
	String targetURL = "";
	String URLParameters = "";
	String header = "";
	int amountOfAllImages = 0;
	int amountOfImages = 0;
	
	//System.out.println("inits ready");
	
	if (!onlyContent) {
	//System.out.println("!onlyContent");	    
	    // Node is open --> "[-] node"
	    if ( isOpen ) {
		    expanderOrCollapserHTML = "[<a class=\"facetOpen\" href=\"/wwwmuseo/servlets/ActionServlet?action=closeClass&facet="+
		    	facet.getId()+
			"&target=hiBrowse&class="+java.net.URLEncoder.encode(node.getId().getString())+
			"\" target=\"_self\">&#150;</a>]";
	    }
		    
	    // Node is closed...
	    else{
		    
		// ...but has content --> "[+] node"
		if (hasContent) {
			//System.out.println("hasContent");
			expanderOrCollapserHTML = "[<a href=\"/wwwmuseo/servlets/ActionServlet?action=openClass&facet="+
				facet.getId()+"&target=hiBrowse&class="+java.net.URLEncoder.encode(node.getId().getString())+
				"\" class=\"facetOpen\">+</a>]";
		}
			
		// doesn't have content --> "node"
		else{
		    //out.println("[&nbsp;]");
		}
	    }
		    
	    // Print node
	    
		 header = node.getLabel();
		 targetURL = "/wwwmuseo/servlets/ActionServlet";
		 URLParameters = "paramNames=selectionPath,facet,class&action=selectClass&selectionPath="+
		 		  java.net.URLEncoder.encode(selectionPath)+"&facet="+
				  facet.getId()+"&target=hiBrowse&class="+
				  java.net.URLEncoder.encode(node.getId().getString());
		 
		 
		//out.println("<a href=\"?+"\" class=\"facetClass\">"++"</a>"); 
		
		amountOfAllImages = (int)node.getImageCount();
		amountOfImages = picsAll.size();
	} 

	String iconURL = ICON_URL+"class.gif";
	
	ClassNodeBean cnb = createClassNodeBean(id, depth, targetURL, URLParameters, iconURL, amountOfImages,
						amountOfAllImages, 0, expanderOrCollapserHTML, header);
	
	
	if(facet.getActiveNode() != null){
		if(cnb.getId().compareTo(facet.getActiveNode().getId().getString()) != 0){
			visibleNodes.add(cnb);
		}else{
			//depth = depth-1;
		}
		
	}else{
		visibleNodes.add(cnb);
	}
	
	
	// if node is open and has content...
	if ( isOpen && hasContent ) {
	//System.out.println("isOpen && hasContent ");
	    // print pictures (if any)
	    Vector temp = new Vector();
	    for (int i = 0; i < pics.size(); ++i) {
		//System.out.println("adding images");
		try{
			Image image = ((Image)pics.elementAt(i));
			
			id = image.getId().getString();
			targetURL = "/wwwmuseo/servlets/ActionServlet";
			URLParameters = "action=setSelectedPicture&class="+java.net.URLEncoder.encode(image.getId().getString())+"&target=main";
			iconURL = ICON_URL+"image.gif";

			Iterator instances = image.getInstancesForClass(node).iterator();
			
			
			
			
			if(facet.getOrderBy() == ORDER_BY_IMAGES){
			
				ImageNodeBean inb = createImageNodeBean(id, depth+1, targetURL, URLParameters, iconURL, "");
			
				while(instances.hasNext()){
					//System.out.println("while---images");
					Instance in = (Instance)instances.next();
					if(in.getLabel() != null){
						inb.addInstanceName(in.getLabel());
					}
				
				}
			
			
				
				temp.add(inb);
			}else{
				
				while(instances.hasNext()){
					//System.out.println("while---instances");
					Instance in = (Instance)instances.next();
					ImageNodeBean inb = createImageNodeBean(id, depth+1, targetURL, URLParameters, iconURL, "");
					if(in.getLabel() != null){
					
						inb.addInstanceName(in.getLabel());
					
						temp.add(inb);
					}
				
				}
				
				
			}
			

			
			
		    //printImage((Image)pics.elementAt(i), node);
		}catch (Exception e){e.printStackTrace();}
	    }
	    
	    if(rootOpen){
		    // print children (if any)
		    for (int i = 0; i < children.size(); ++i) {
			    //System.out.println("adding children");
			    OntologyClass child = (OntologyClass)children.elementAt(i);
			    addNodeBean(visibleNodes, facet, selectionPath + Profile.ONTOLOGY_CLASS_SEPARATOR + node.getId().getString(), child, false, depth+1, true);
		    }
	    }
	    
	    
			//System.out.println("start short:");
			Collections.sort(temp);
			//System.out.println("end short - start add all");
			visibleNodes.addAll(temp);	    

	    

	}
    }

    
    /*
     * Helper method used to creat ImageNdeBeans
     */
    private static ImageNodeBean createImageNodeBean(String id, int depth, String targetURL,
    	String URLParameters, String iconURL,String instanceNames){
	
	    ImageNodeBean nb = new ImageNodeBean();
	    nb.setId(id);
	    nb.setDepth(depth);
	    nb.setTargetURL(targetURL);
	    nb.setURLParameters(URLParameters);
	    nb.setIconURL(iconURL);
	    //nb.setInstanceNames(instanceNames);
	    
	    return nb;
    }    
    
   
    /*
     * Helper method used to create ClassNodeBeans
     */
    private static ClassNodeBean createClassNodeBean(String id, int depth, String targetURL,
    	String URLParameters, String iconURL,
	int amountOfImages, int amountOfAllImages, int amountOfInstances,
	String expanderOrCollapserHTML, String header){
		
	//System.out.println("createClassNodeBean");
	    ClassNodeBean nb = new ClassNodeBean();
	    nb.setId(id);
	    nb.setDepth(depth);
	    nb.setTargetURL(targetURL);
	    nb.setURLParameters(URLParameters);
	    nb.setIconURL(iconURL);
	    nb.setAmountOfImages(amountOfImages);
	    nb.setAmountOfAllImages(amountOfAllImages);
	    nb.setAmountOfInstances(amountOfInstances);
	    nb.setExpanderOrCollapserHTML(expanderOrCollapserHTML);
	    nb.setHeader(header);
	    
	    return nb;
    }
    
    /**  
     * Returns all visible facets
     * @return Vector of FacetSelections. 
     */
    public Vector getStateOfHiBrowse(){
	//System.out.println("return state of hibrowse");
	return this.selectedFacets;
    }
	
	
    /**
     * @return Vector containing Strings. These are all previous selections of the current user.
     * Vector's elements are in the same order than the user's selections. elementAt(0) is the latest
     * and the elementAt(n) is the first;
     */
	
    public Vector getPreviousImages() {
	return this.previousSelections;
    }
	
    
    /**
     * Adds one image id to the previously shown images
     */
    public void setSelectedPicture(String facet, String id){
	model.getImage(new SimpleId(id)).addWeight();
	this.previousSelections.add(new SimpleId(id));
    }


    /**
     * Removes one class from opened classes list
     */
    public void closeClass( String facet, String oclass ) {
	//System.out.println( "closeClass("+facet+", "+oclass+")");

	try{
	    int i = Collections.binarySearch( this.selectedFacets, new FacetSelection(Integer.parseInt(facet)) );
	    if (i >= 0)
		((FacetSelection)selectedFacets.elementAt(i)).removeOpened(oclass);
	} catch (NumberFormatException nfe) {}
    }


    /**
     * Removes facet from shown facets list
     */
    public void killFacet( String facet ) {
	//System.out.println( "killFacet("+facet+", "+oclass+")");
	try {
	    int i = Collections.binarySearch( this.selectedFacets, new FacetSelection(Integer.parseInt(facet)) );
	    if (i >= 0)
		this.selectedFacets.remove(i);
	} catch (NumberFormatException nfe) {}
	
    }
    

    /**
     * Adds class to the opened classes list
     */
    public void openClass( String facet, String oclass ) {
	//System.out.println( "openClass("+facet+", "+oclass+")");
	
	try{
	    int i = Collections.binarySearch( this.selectedFacets, new FacetSelection(Integer.parseInt(facet)) );
	    if (i >= 0)
		((FacetSelection)selectedFacets.elementAt(i)).addOpen(oclass);
	} catch (NumberFormatException nfe) {}
    }


    /**
     * Sets the class selected from the facet.
     */
    public void selectClass( String selectionPath, String facet, String oclass ) {
	//System.out.println( "selectClass("+selectionPath+", "+facet+", "+oclass+")");
	openClass(facet, oclass);
	
	try{
	    int i = Collections.binarySearch( this.selectedFacets, new FacetSelection(Integer.parseInt(facet)) );
	    if (i >= 0) {
		FacetSelection fs = (FacetSelection)selectedFacets.elementAt(i);
		
		if(oclass.compareTo("deSelect") == 0){
			fs.setSelectionPath("");
			fs.setActiveNode(null);
		}else{
			fs.setSelectionPath(selectionPath);
			fs.setActiveNode(model.getOntologyClass(new SimpleId(oclass)));
		}
	    }
	} catch (NumberFormatException nfe) {}
    }
    

    /**
     * Adds the facet to the visible facet list
     */
    public void showFacet( String root ) {
	filter = W3m.getFilter();
	OntologyClass fsRoot = model.getOntologyClass(new SimpleId(root));
	//System.out.println( "showFacet("+facet+")");
	if (root != null)
	    this.selectedFacets.add( new FacetSelection( getNewFacetId(), fsRoot ) );
    }

    /**
     * Sets the way of ordering this facets images.
     * order should be "instances" if the facet should be ordered by instances
     * othervise thi facet is ordered by images.
     */
    
    public void setOrderBy(String order, String facet){
	try{
	    int i = Collections.binarySearch( this.selectedFacets, new FacetSelection(Integer.parseInt(facet)) );
	    if (i >= 0) {
		FacetSelection fs = (FacetSelection)selectedFacets.elementAt(i);
		
		if(order.compareTo("instances") == 0){
			fs.setOrderBy(ORDER_BY_INSTANCES);
			System.out.println("ORDER_BY_INSTANCES");
		}else{
			fs.setOrderBy(ORDER_BY_IMAGES);
			System.out.println("ORDER_BY_IMAGES");
		}
	    }
	} catch (NumberFormatException nfe) {}
    }
}
